package de.ullisroboterseite.ursai2applauncher;

// Autor: http://UllisRoboterSeite.de

// Doku:  http://UllisRoboterSeite.de/android-AI2-AppLauncher.html
// Created: 2020-04-16
//
// Version 1 (2020-04-16)
// -------------------------
// - Basis-Version
//
// Version 2 (2020-04-29)
// -------------------------
// - Wertet Rückgaben der gestarteten Activity aus
// - Vereinfachter Launch von AI2 Apps
// - Unterscheidet zwischen SubActivity und neuem Task
//
// Version 2.1 (2020-06-03)
// -------------------------
// - MyAppName, MyPackageName hinzugefügt
//
// Version 2.2 (2021-04-23)
// -------------------------
// - Permission SYSTEM_ALERT_WINDOW & ACTION_MANAGE_OVERLAY_PERMISSION hinzugefügt.
// - Eigenschaft VersionName hinzugefügt
// - Eigenschaft VersionSDK hinzugefügt
// - Eigenschaft Manufacturer hinzugefügt
// - Eigenschaft IsXiaomi hinzugefügt
// - Eigenschaft CanLaunchFromBackground hinzugefügt
// - Eigenschaft CanDrawOverlays hinzugefügt
// - Methode CheckBackgroundLaunchPermission hinzugefügt
// - Methode OpenMiuiPermissionEditor hinzugefügt
// - Ereignis AfterBackgroundPermssion hinzugefügt
// - Eigenschaft PermissionDelay hinzugefügt
// - Extension LauncherIntent hinzugefügt
// - Methoden LaunchAppFromIntent & LaunchForResultFromIntent hinzugefügt
// - Eigenschaft CurrentLanguage hinzugefügt
//
//
// Version 2.3 (2021-05-12)
// ------------------------
// - ScreenToOpen hat bei Kodular nicht funktionert. Context.getPackageName() ist nicht nutzbar.
//   stattdessen: form.getClass().getPackage().getName()
//
// Version 2.4 (2021-05-15)
// ------------------------
// - ScreenToOpen hat bei Kodular nicht funktionert.
//
// Version 2.5 (2021-06-13)
// ------------------------
// - flagNewTask wurde nicht korrekt ausgewertet, ist aber für Android 10 notwendig.
// - Default für flagNewTask ist true.
//
// Version 2.6 (2023-09-10)
// ------------------------
// - Ab Android 11 ist die Permission QUERY_ALL_PACKAGES notwendig, wenn man alle Dateien in der Liste sehen will.


import java.util.*;

import com.google.appinventor.components.annotations.*;
import com.google.appinventor.components.annotations.androidmanifest.*;
import com.google.appinventor.components.common.*;
import com.google.appinventor.components.runtime.*;
import com.google.appinventor.components.runtime.util.*;

import android.content.*;
import android.content.pm.*;
import android.util.Log;
import android.app.Activity;

import android.os.Handler;
import android.os.Bundle;
import android.os.Build;

import android.provider.Settings;
import android.net.Uri;

@DesignerComponent(version = 2, //
        versionName = UrsAI2AppLauncher.VersionName, //
        dateBuilt = UrsAI2AppLauncher.dateBuilt, //
        description = "AI2 extension block for app launching.", //
        category = com.google.appinventor.components.common.ComponentCategory.EXTENSION, //
        nonVisible = true, //
        helpUrl = "http://UllisRoboterSeite.de/android-AI2-AppLauncher.html", //
        iconName = "aiwebres/icon.png")
@SimpleObject(external = true)
@UsesPermissions(permissionNames = "android.permission.SYSTEM_ALERT_WINDOW, android.permission.ACTION_MANAGE_OVERLAY_PERMISSION, android.permission.QUERY_ALL_PACKAGES")
public class UrsAI2AppLauncher extends AndroidNonvisibleComponent implements ActivityResultListener {
    static final String LOG_TAG = "LAUNCH";
    static final String VersionName = "2.6.0";
    static final String dateBuilt = "2023-09-10";

    private final UrsAI2AppLauncher thisInstance = this;
    private final Handler handler = new Handler();

    private final AppLauncherListUtil appLauncherListUtil;
    private final AppLauncherUtil appLauncherUtil;

    private final int launcherRequestCode;
    private final int permissionRequestCode;
    private static final String RESULT_NAME = "APP_INVENTOR_RESULT";
    private static final String ARGUMENT_NAME = "APP_INVENTOR_START";

    private Bundle lastResultDataBundle = null;

    public UrsAI2AppLauncher(ComponentContainer container) {
        super(container.$form());
        appLauncherListUtil = new AppLauncherListUtil(form);
        launcherRequestCode = form.registerForActivityResult(this);
        permissionRequestCode = form.registerForActivityResult(this);
        appLauncherUtil = new AppLauncherUtil(form, launcherRequestCode);
    } // ctor

    @SimpleProperty(description = "Returns the extension's version name.")
    public String Version() {
        return VersionName;
    }

    @SimpleProperty(description = "Returns running Android SDK version (API level).")
    public int VersionSDK() {
        return Build.VERSION.SDK_INT;
    }

    @SimpleProperty(description = "Returns the manufacturer name of the device.")
    public String Manufacturer() {
        return Build.MANUFACTURER;
    }

    @SimpleProperty(description = "Returns whether this is a Xiaomi device.")
    public boolean IsXiaomi() {
        return "xiaomi".equals(Build.MANUFACTURER.toLowerCase(Locale.ROOT));
    }

    // https://de.wikipedia.org/wiki/Liste_der_ISO-639-1-Codes
    // https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes
    @SimpleProperty(description = "Returns the current language, e.g. 'de' or 'en'.")
    public String CurrentLanguage() {
        return Locale.getDefault().getLanguage();
    }

    // #region Paket- und Namenslisten =================================
    @SimpleFunction(description = "Returns a list of Strings containing the package names of all installed apps.\n"
            + "If Filter != \"\" all apps are listed otherwise those with package names containing the specified string. Ignores cases.")
    public YailList getPackageList(String Filter) {
        return appLauncherListUtil.getPackageList(Filter);
    }

    @SimpleFunction(description = "Returns a list of Strings containing the names of all installed apps.\n"
            + "If Filter != \"\" all apps are listed otherwise those with names containing the specified string. Ignores cases.")
    public YailList getNameList(String Filter) {
        return appLauncherListUtil.getNameList(Filter);
    }

    @SimpleFunction(description = "Gets the package name from the specified app name. Returns \"\" if not found.")
    public String PackageFromName(String AppName) {
        return appLauncherListUtil.PackageFromName(AppName);
    }
    // #endregion

    // #region Launch App ==============================================
    @SimpleEvent(description = "Result of app launching.\n"
            + "Result: 0: success, 1: app not found, 2: invalid extra data list.")
    public void AfterAppLaunched(int Result) {
        EventDispatcher.dispatchEvent(this, "AfterAppLaunched", Result);
    }

    @SimpleFunction(description = "Launches the specified app.")
    public void LaunchApp(String PackageName) {
        lastResultDataBundle = null;
        AfterAppLaunched(appLauncherUtil.launchApp(PackageName, true));
    }

    @SimpleFunction(description = "Launches the specified AI2 app.")
    public void LaunchAI2App(String PackageName, String StartValue) {
        lastResultDataBundle = null;
        AfterAppLaunched(appLauncherUtil.launchAppExtra(PackageName, ARGUMENT_NAME, StartValue, true));
    }

    @SimpleFunction(description = "Launches the specified app. Key and Value must not be empty.")
    public void LaunchAppExtra(String PackageName, String Key, String Value) {
        lastResultDataBundle = null;
        AfterAppLaunched(appLauncherUtil.launchAppExtra(PackageName, Key, Value, true));
    }

    @SimpleFunction(description = "Launches the specified app.")
    public void LaunchAppExtraList(String PackageName, YailList Pairs) {
        lastResultDataBundle = null;
        AfterAppLaunched(appLauncherUtil.launchAppExtraList(PackageName, Pairs, true));
    }

    @SimpleFunction(description = "Launches the specified app.")
    public void LaunchAppFromIntent(Component LauncherIntentObject) {
        if (!(LauncherIntentObject instanceof LauncherIntent)) {
            AfterAppLaunched(3);
            return;
        }

        lastResultDataBundle = null;
        Intent intent = ((LauncherIntent) LauncherIntentObject).buildIntent();
        AfterAppLaunched(appLauncherUtil.launchAppFromIntent(intent));
    }

    // ------------------------------------------------------

    @SimpleFunction(description = "Launches the specified app and waits for the app to return a value.")
    public void LaunchForResult(String PackageName) {
        lastResultDataBundle = null;
        AfterAppLaunched(appLauncherUtil.launchApp(PackageName, false));
    }

    @SimpleFunction(description = "Launches the specified app and waits for the app to return a value..")
    public void LaunchForResultAI2App(String PackageName, String StartValue) {
        lastResultDataBundle = null;
        AfterAppLaunched(appLauncherUtil.launchAppExtra(PackageName, ARGUMENT_NAME, StartValue, false));
    }

    @SimpleFunction(description = "Launches the specified app and waits for the app to return a value. Key and Value must not be empty.")
    public void LaunchForResultAppExtra(String PackageName, String Key, String Value) {
        lastResultDataBundle = null;
        AfterAppLaunched(appLauncherUtil.launchAppExtra(PackageName, Key, Value, false));
    }

    @SimpleFunction(description = "Launches the specified app and waits for the app to return a value.")
    public void LaunchForResultAppExtraList(String PackageName, YailList Pairs) {
        lastResultDataBundle = null;
        AfterAppLaunched(appLauncherUtil.launchAppExtraList(PackageName, Pairs, false));
    }

    @SimpleFunction(description = "Launches the specified app.")
    public void LaunchForResultFromIntent(Component LauncherIntentObject) {
        if (!(LauncherIntentObject instanceof LauncherIntent)) {
            AfterAppLaunched(3);
            return;
        }

        lastResultDataBundle = null;
        Intent intent = ((LauncherIntent) LauncherIntentObject).buildIntent();
        AfterAppLaunched(appLauncherUtil.launchAppFromIntentForResult(intent));
    }
    // #endregion

    // #region Activity Result =========================================

    @SimpleEvent(description = "Event raised after called App returns.")
    public void AfterActivity(String Result) {
        EventDispatcher.dispatchEvent(this, "AfterActivity", Result);
    }

    @SimpleEvent(description = "Event raised if the called App was canceled.")
    public void ActivityCanceled() {
        EventDispatcher.dispatchEvent(this, "ActivityCanceled");
    }

    /**
     * The callback method used to report Activity results back to the caller.
     *
     * @param requestCode the originally passed in request code. Used to identify
     *                    the call.
     * @param resultCode  the returned result code:
     *                    {@link android.app.Activity#RESULT_OK} or
     *                    {@link android.app.Activity#RESULT_CANCELED}
     * @param data        the returned data, encapsulated as an {@link Intent}.
     */
    @Override
    public void resultReturned(int requestCode, int resultCode, Intent data) {
        if (requestCode == launcherRequestCode)
            handleLaucherResult(requestCode, resultCode, data);

        if (requestCode == permissionRequestCode)
            AfterBackgroundPermssion(true); // Wert is egal
    }

    void handleLaucherResult(int requestCode, int resultCode, Intent data) {
        String result = "";
        if (resultCode == Activity.RESULT_CANCELED) {
            ActivityCanceled();
            return;
        }

        // resultCode == Activity.RESULT_OK)
        if (data != null) {
            lastResultDataBundle = data.getExtras();
            if (lastResultDataBundle != null) {
                if (lastResultDataBundle.get(RESULT_NAME) != null)
                    result = "" + lastResultDataBundle.get(RESULT_NAME);
            }
        }

        // call user's AfterActivity event handler
        AfterActivity(result);
    }

    @SimpleFunction(description = "Returns a list with all key names the launched app returns.")
    public YailList getResultKeys(String Filter) {
        List<String> keyList = new ArrayList<String>();
        if (lastResultDataBundle != null) {
            for (String key : lastResultDataBundle.keySet()) {
                if (Filter.isEmpty())
                    keyList.add(key);
                else if (key.contains(Filter))
                    keyList.add(key);
            }
        }
        Collections.sort(keyList);

        return YailList.makeList(keyList);
    }

    @SimpleFunction(description = "Returns the value that corresponds to the key name.")
    public String getResultValue(String Key, String Default) {
        if (lastResultDataBundle == null)
            return Default;
        if (lastResultDataBundle.get(Key) == null)
            return Default;

        try {
            String value = "" + lastResultDataBundle.get(Key);
            return value;
        } catch (Exception e) {
            return Default;
        }
    }

    @SimpleFunction(description = "Returns the object that corresponds to the key name. null if not available.")
    public Object getResultObject(String Key) {
        if (lastResultDataBundle == null)
            return null;

        return lastResultDataBundle.get(Key);
    }
    // #endregion

    // #region StartValues =============================================
    @SimpleFunction(description = "Returns a list with all key names the calling app provides.")
    public YailList getStartKeys(String Filter) {
        List<String> keyList = new ArrayList<String>();
        Intent startIntent = form.getIntent();

        if (startIntent != null) {
            Bundle bundle = startIntent.getExtras();
            if (bundle != null) {
                for (String key : bundle.keySet()) {
                    if (Filter.isEmpty())
                        keyList.add(key);
                    else if (key.contains(Filter))
                        keyList.add(key);
                }
            }
        }
        Collections.sort(keyList);

        return YailList.makeList(keyList);
    }

    @SimpleFunction(description = "Returns the value that corresponds to the key name.")
    public String getStartValue(String Key, String Default) {
        Intent startIntent = form.getIntent();

        if (startIntent != null)
            return Default;

        Bundle bundle = startIntent.getExtras();

        if (bundle == null)
            return Default;

        if (bundle.get(Key) == null)
            return Default;

        try {
            String value = "" + bundle.get(Key);
            return value;
        } catch (Exception e) {
            return Default;
        }
    }

    @SimpleFunction(description = "Returns the object that corresponds to the key name. null if not available.")
    public Object getStartObject(String Key) {
        Intent startIntent = form.getIntent();

        if (startIntent != null)
            return null;

        Bundle bundle = startIntent.getExtras();

        if (bundle == null)
            return null;

        return lastResultDataBundle.get(Key);
    }
    // #endregion

    // #region App Info ================================================
    @SimpleProperty(description = "Returns the name of this app.")
    public String MyAppName() {
        try {
            return form.getApplicationInfo().loadLabel(form.getPackageManager()).toString();
        } catch (Exception e) {
            return "";
        }
    }

    @SimpleProperty(description = "Returns the package name of this app.")
    public String MyPackageName() {
        return form.getPackageName();
    }
    // #endregion

    // #region Overlay Permission ================================================

    int permissionDelay = 500;
    int tempPermissionDelay = permissionDelay;

    @DesignerProperty(editorType = PropertyTypeConstants.PROPERTY_TYPE_NON_NEGATIVE_INTEGER, defaultValue = "500")
    @SimpleProperty(description = "Delay for requesting CanLaunchFromBackground after CheckBackgroundLaunchPermission.")
    public void PermissionDelay(int value) {
        permissionDelay = value;
    }

    @SimpleProperty(description = "An identifier for this notification unique within the application.")
    public int PermissionDelay() {
        return permissionDelay;
    }

    @SimpleProperty(description = "Returns whether permission SYSTEM_ALERT_WINDOW is granted.")
    public boolean CanDrawOverlays() {
        if (Build.VERSION.SDK_INT < 23)
            return true;
        else
            return Settings.canDrawOverlays(form);
    }

    @SimpleProperty(description = "Returns whether you are permitted to lauch apps while in background.")
    public boolean CanLaunchFromBackground() {
        if (Build.VERSION.SDK_INT < 29)
            return true;
        else
            return Settings.canDrawOverlays(form);
    }

    @SimpleFunction(description = "Checks if lauching apps from the background is permitted and if not, asks for the permission.")
    public void CheckBackgroundLaunchPermission() {
        if (Build.VERSION.SDK_INT < 29) { // Nicht notwendig für Versionen < Android 10
            tempPermissionDelay = 1;
            AfterBackgroundPermssion(true);
            return;
        }

        if (CanLaunchFromBackground()) { // Erlaubnis bereits erteilt?
            tempPermissionDelay = 1;
            AfterBackgroundPermssion(true);
            return;
        }

        tempPermissionDelay = permissionDelay;
        Intent overlaySettings = new Intent(Settings.ACTION_MANAGE_OVERLAY_PERMISSION,
                Uri.parse("package:" + form.getPackageName()));
        form.startActivityForResult(overlaySettings, permissionRequestCode);
    }

    @SimpleEvent(description = "Returns the result of CheckBackgroundLaunchPermission.")
    public void AfterBackgroundPermssion(boolean Result) {
        final UrsAI2AppLauncher thisInstance = this;
        final Form f = form;
        handler.postDelayed(new Runnable() { // Das Ergebnis steht erst nach einer gewissen Zeit zur Verfügung.
            public void run() {
                EventDispatcher.dispatchEvent(thisInstance, "AfterBackgroundPermssion", CanLaunchFromBackground());
            }
        }, tempPermissionDelay);
    }

    @SimpleFunction(description = "Opens the MIUI permission editor.")
    public void OpenMiuiPermissionEditor() {
        if ("xiaomi".equals(Build.MANUFACTURER.toLowerCase(Locale.ROOT))) {
            final Intent intent = new Intent("miui.intent.action.APP_PERM_EDITOR");
            intent.setClassName("com.miui.securitycenter", "com.miui.permcenter.permissions.PermissionsEditorActivity");
            intent.putExtra("extra_pkgname", form.getPackageName());
            form.startActivity(intent);
        }
    }

    // #endregion
}